@@ -1,6 +1,6 @@
 /**
  *
- * Copyright (c) 2013-2018 Pascal Gauthier.
+ * Copyright (c) 2013-2025 Pascal Gauthier.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -31,6 +31,7 @@
 #include "msfa/exp2.h"
 #include "msfa/env.h"
 #include "msfa/pitchenv.h"
+#include "msfa/porta.h"
 #include "msfa/aligned_buf.h"
 #include "msfa/fm_op_kernel.h"
 
@@ -66,7 +67,6 @@
 DexedAudioProcessor::DexedAudioProcessor()
     : AudioProcessor(BusesProperties().withOutput("output", AudioChannelSet::stereo(), true)) {
 #ifdef DEBUG
-    
     // avoid creating the log file if it is in standalone mode
     if ( !JUCEApplication::isStandaloneApp() ) {
         Logger *tmp = Logger::getCurrentLogger();
@@ -90,7 +90,7 @@ DexedAudioProcessor::DexedAudioProcessor()
     
     vuSignal = 0;
     monoMode = 0;
-    
+
     resolvAppDir();
     
     initCtrl();
@@ -136,24 +136,30 @@ void DexedAudioProcessor::prepareToPlay(double sampleRate, int samplesPerBlock)
     Lfo::init(sampleRate);
     PitchEnv::init(sampleRate);
     Env::init_sr(sampleRate);
+    Porta::init_sr(sampleRate);
     fx.init(sampleRate);
 
     vuDecayFactor = VuMeterOutput::getDecayFactor(sampleRate);
     
     for (int note = 0; note < MAX_ACTIVE_NOTES; ++note) {
         voices[note].dx7_note = new Dx7Note(synthTuningState, mtsClient);
+        voices[note].midi_note = -1;
         voices[note].keydown = false;
         voices[note].sustained = false;
         voices[note].live = false;
+        voices[note].keydown_seq = -1;
     }
 
     currentNote = 0;
+    nextKeydownSeq = 0;
     controllers.values_[kControllerPitch] = 0x2000;
     controllers.modwheel_cc = 0;
     controllers.foot_cc = 0;
     controllers.breath_cc = 0;
     controllers.aftertouch_cc = 0;
-	controllers.refresh(); 
+    controllers.portamento_enable_cc = false;
+    controllers.portamento_cc = 0;
+	controllers.refresh();
 
     sustain = false;
     extra_buf_size = 0;
@@ -353,14 +359,15 @@ void DexedAudioProcessor::processMidiMessage(const MidiMessage *msg) {
         {
             int i = voiceIndex;
             switch(cf0) {
-            case 0xb0:
-                voices[i].mpeTimbre = (int)buf[2];
-                voices[i].dx7_note->mpeTimbre = (int)buf[2];
-                break;
-            case 0xd0:
-                voices[i].mpePressure = (int)buf[1];
-                voices[i].dx7_note->mpePressure = (int)buf[1];
-                break;
+            // THIS IS COMMENTED SINCE mepTimbre and mpePressure is not used
+            // case 0xb0:
+            //     voices[i].mpeTimbre = (int)buf[2];
+            //     voices[i].dx7_note->mpeTimbre = (int)buf[2];
+            //     break;
+            // case 0xd0:
+            //     voices[i].mpePressure = (int)buf[1];
+            //     voices[i].dx7_note->mpePressure = (int)buf[1];
+            //     break;
             case 0xe0:
                 voices[i].mpePitchBend = (int)( buf[1] | (buf[2] << 7) );
                 voices[i].dx7_note->mpePitchBend = (int)( buf[1] | ( buf[2] << 7 ) );
@@ -384,8 +391,6 @@ void DexedAudioProcessor::processMidiMessage(const MidiMessage *msg) {
         case 0xb0 : {
             int ctrl = buf[1];
             int value = buf[2];
-
-
                 switch(ctrl) {
                 case 1:
                     controllers.modwheel_cc = value;
@@ -399,6 +404,9 @@ void DexedAudioProcessor::processMidiMessage(const MidiMessage *msg) {
                     controllers.foot_cc = value;
                     controllers.refresh();
                     break;
+                case 5:
+                    controllers.portamento_cc = value;
+                    break;
                 case 64:
                     sustain = value > 63;
                     if (!sustain) {
@@ -410,6 +418,9 @@ void DexedAudioProcessor::processMidiMessage(const MidiMessage *msg) {
                         }
                     }
                     break;
+                case 65:
+                    controllers.portamento_enable_cc = value >= 64;
+                    break;
                 case 120:
                     panic();
                     break;
@@ -420,16 +431,17 @@ void DexedAudioProcessor::processMidiMessage(const MidiMessage *msg) {
                     }
                     break;
                 default:
-                    TRACE("handle CC %d %d", ctrl, value);
-                    if ( mappedMidiCC.contains(ctrl) ) {
-                        Ctrl *linkedCtrl = mappedMidiCC[ctrl];
+                    TRACE("handle channel %d CC %d = %d", channel, ctrl, value);
+                    int channel_cc = (channel << 8) | ctrl;
+                    if ( mappedMidiCC.contains(channel_cc) ) {
+                        Ctrl *linkedCtrl = mappedMidiCC[channel_cc];
                         
                         // We are not publishing this in the DSP thread, moving that in the
                         // event thread
                         linkedCtrl->publishValueAsync((float) value / 127);
                     }
                     // this is used to notify the dialog that a CC value was received.
-                    lastCCUsed.setValue(ctrl);
+                    lastCCUsed.setValue(channel_cc);
                 }
             }
             return;
@@ -452,6 +464,31 @@ void DexedAudioProcessor::processMidiMessage(const MidiMessage *msg) {
 
 #define ACT(v) (v.keydown ? v.midi_note : -1)
 
+int DexedAudioProcessor::chooseNote(uint8_t pitch) {
+    // order of preference:
+    // 1. a note that is not playing
+    // 2. a note with its key up, playing the same pitch
+    // 3. a note with its key up, playing a different pitch
+    // 4. a note with its key down, playing the same pitch
+    // 5. a note with its key down, playing a different pitch
+    // break ties by preferring note with least recent keydown
+    int bestNote = currentNote;
+    int bestScore = -1;
+    int note = currentNote;
+    for (int i=0; i<MAX_ACTIVE_NOTES; i++) {
+        int score = 0;
+        if ( !voices[note].dx7_note->isPlaying() ) score += 4;
+        if ( !voices[note].keydown ) score += 2;
+        if ( voices[note].midi_note == pitch ) score += 1;
+        if ( (score > bestScore) || (score == bestScore && voices[note].keydown_seq < voices[bestNote].keydown_seq) ) {
+            bestNote = note;
+            bestScore = score;
+        }
+        note = (note + 1) % MAX_ACTIVE_NOTES;
+    }
+    return bestNote;
+}
+
 void DexedAudioProcessor::keydown(uint8_t channel, uint8_t pitch, uint8_t velo) {
     if ( velo == 0 ) {
         keyup(channel, pitch, velo);
@@ -463,12 +500,11 @@ void DexedAudioProcessor::keydown(uint8_t channel, uint8_t pitch, uint8_t velo)
     if ( normalizeDxVelocity ) {
         velo = ((float)velo) * 0.7874015; // 100/127
     }
+  
 
-    if( controllers.mpeEnabled )
-    {
+    if( controllers.mpeEnabled ) {
         int note = currentNote;
-        for( int i=0; i<MAX_ACTIVE_NOTES; ++i )
-        {
+        for( int i=0; i<MAX_ACTIVE_NOTES; ++i ) {
             if( voices[note].keydown && voices[note].channel == channel )
             {
                 // If we get two keydowns on the same channel we are getting information from a non-mpe device
@@ -477,25 +513,38 @@ void DexedAudioProcessor::keydown(uint8_t channel, uint8_t pitch, uint8_t velo)
             note = (note + 1) % MAX_ACTIVE_NOTES;
         }
     }
-    
-    int note = currentNote;
+
+    bool triggerLfo = true;
     for (int i=0; i<MAX_ACTIVE_NOTES; i++) {
-        if (!voices[note].keydown) {
-            currentNote = (note + 1) % MAX_ACTIVE_NOTES;
-            lfo.keydown();  // TODO: should only do this if # keys down was 0
-            voices[note].channel = channel;
-            voices[note].midi_note = pitch;
-            voices[note].velocity = velo;
-            voices[note].sustained = sustain;
-            voices[note].keydown = true;
-            voices[note].dx7_note->init(data, pitch, velo, channel);
-            if ( data[136] )
-                voices[note].dx7_note->oscSync();
+        if ( voices[i].keydown ) {
+            triggerLfo = false;
             break;
         }
-        note = (note + 1) % MAX_ACTIVE_NOTES;
     }
-    
+    if ( triggerLfo ) {
+        lfo.keydown();
+    }
+
+    int note = chooseNote(pitch);
+
+    currentNote = (note + 1) % MAX_ACTIVE_NOTES;
+    voices[note].channel = channel;
+    voices[note].midi_note = pitch;
+    voices[note].velocity = velo;
+    voices[note].sustained = sustain;
+    voices[note].keydown = true;
+    voices[note].keydown_seq = nextKeydownSeq++;
+    // to avoid click, don't sync oscillators when voice stealing
+    bool voice_steal = voices[note].dx7_note->isPlaying();
+    voices[note].dx7_note->init(data, pitch, velo, channel, &controllers);
+    if ( data[136] && !voice_steal ) {
+        voices[note].dx7_note->oscSync();
+    }
+    if ( (voices[lastActiveVoice].midi_note != -1 && controllers.portamento_enable_cc)
+       && controllers.portamento_cc > 0 ) {
+        voices[note].dx7_note->initPortamento(*voices[lastActiveVoice].dx7_note);
+    }
+
     if ( monoMode ) {
         for(int i=0; i<MAX_ACTIVE_NOTES; i++) {            
             if ( voices[i].live ) {
@@ -514,8 +563,22 @@ void DexedAudioProcessor::keydown(uint8_t channel, uint8_t pitch, uint8_t velo)
             }
         }
     }
+    else if ( !data[136] ) {
+        // if another note at the same pitch is playing, transfer phase
+        // to avoid unpredictable destructive interference. this can cause
+        // clicking when voice stealing, but we've tried to choose a voice
+        // to steal that will minimise the chances of clicking
+        for(int i=0; i<MAX_ACTIVE_NOTES; i++) {
+            if ( i != note && voices[i].dx7_note->isPlaying() && voices[i].midi_note == pitch ) {
+                voices[note].dx7_note->transferPhase(*voices[i].dx7_note);
+                break;
+            }
+        }
+    }
+
  
     voices[note].live = true;
+    lastActiveVoice = note;
 	//TRACE("activate %d [ %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d ]", pitch, ACT(voices[0]), ACT(voices[1]), ACT(voices[2]), ACT(voices[3]), ACT(voices[4]), ACT(voices[5]), ACT(voices[6]), ACT(voices[7]), ACT(voices[8]), ACT(voices[9]), ACT(voices[10]), ACT(voices[11]), ACT(voices[12]), ACT(voices[13]), ACT(voices[14]), ACT(voices[15]));
 }
 
@@ -584,6 +647,7 @@ int DexedAudioProcessor::tuningTranspositionShift()
 
 void DexedAudioProcessor::panic() {
     for(int i=0;i<MAX_ACTIVE_NOTES;i++) {
+        voices[i].midi_note = -1;
         voices[i].keydown = false;
         voices[i].live = false;
         if ( voices[i].dx7_note != NULL ) {
@@ -608,65 +672,84 @@ void DexedAudioProcessor::handleIncomingMidiMessage(MidiInput* source, const Mid
 
     if ( ! message.isSysEx() )
         return;
-    
+
     // test if it is a Yamaha Sysex
     if ( buf[1] != 0x43 ) {
         TRACE("not a yamaha sysex %d", buf[1]);
         return;
     }
-    
+
     int substatus = buf[2] >> 4;
-    
-    if ( substatus == 0 ) {
-        // single voice dump
-        if ( buf[3] == 0 ) {
-            if ( sz < 156 ) {
-                TRACE("wrong single voice datasize %d", sz);
-                return;
+    switch(substatus) {
+        case 0 : {
+            // single voice dump
+            if ( buf[3] == 0 ) {
+                if ( sz < 156 ) {
+                    TRACE("wrong single voice datasize %d", sz);
+                    return;
+                }
+                if ( updateProgramFromSysex(buf+6) )
+                    TRACE("bad checksum when updating program from sysex message");
             }
-            if ( updateProgramFromSysex(buf+6) )
-                TRACE("bad checksum when updating program from sysex message");
-        }
-        
-        // 32 voice dump
-        if ( buf[3] == 9 ) {
-            if ( sz < 4104 ) {
-                TRACE("wrong 32 voice dump data size %d", sz);
-                return;
-            }
-            
-            Cartridge received;
-            if ( received.load(buf, sz) == 0 ) {
-                loadCartridge(received);
-                setCurrentProgram(0);
+
+            // 32 voice dump
+            if ( buf[3] == 9 ) {
+                if ( sz < 4104 ) {
+                    TRACE("wrong 32 voice dump data size %d", sz);
+                    return;
+                }
+
+                Cartridge received;
+                if ( received.load(buf, sz) == 0 ) {
+                    loadCartridge(received);
+                    setCurrentProgram(0);
+                }
             }
         }
-    } else if ( substatus == 1 ) {
-        // parameter change
-        if ( sz < 7 ) {
-           TRACE("wrong single voice datasize %d", sz);
-           return;
-        }
-        
-        uint8 offset = (buf[3] << 7) + buf[4];
-        uint8 value = buf[5];
-        
-        TRACE("parameter change message offset:%d value:%d", offset, value);
-        
-        if ( offset > 155 ) {
-            TRACE("wrong offset size");
+        break;
+        case 1 : {
+            // parameter change
+            if ( sz < 7 ) {
+            TRACE("wrong single voice datasize %d", sz);
             return;
+            }
+
+            uint8 offset = (buf[3] << 7) + buf[4];
+            uint8 value = buf[5];
+
+            TRACE("parameter change message offset:%d value:%d", offset, value);
+
+            if ( offset > 155 ) {
+                TRACE("wrong offset size");
+                return;
+            }
+
+            if ( offset == 155 ) {
+                unpackOpSwitch(value);
+            } else {
+                data[offset] = value;
+            }
         }
-        
-        if ( offset == 155 ) {
-            unpackOpSwitch(value);
-        } else {
-            data[offset] = value;
+        break;
+        case 2: {
+            if ( buf[3] == 0 ) {
+                // single voice request
+                sendCurrentSysexProgram();
+            } else if ( buf[3] == 9 ) {
+                // cart request
+                sendCurrentSysexCartridge();
+            } else {
+                TRACE("Unknown voice request: %d", buf[3]);
+            }
         }
-    } else {
-        TRACE("unknown sysex substatus: %d", substatus);
+        return;
+
+        default: {
+            TRACE("unknown sysex substatus: %d", substatus);
+        }
+        return;
     }
-    
+
     forceRefreshUI = true;
     triggerAsyncUpdate();
 }
@@ -677,7 +760,7 @@ int DexedAudioProcessor::getEngineType() {
 
 void DexedAudioProcessor::setEngineType(int tp) {
     TRACE("settings engine %d", tp);
-    
+
     switch (tp)  {
         case DEXED_ENGINE_MARKI:
             controllers.core = &engineMkI;
@@ -786,33 +869,15 @@ void DexedAudioProcessor::updateUI() {
 }
 
 AudioProcessorEditor* DexedAudioProcessor::createEditor() {
-    static const uint8_t HIGH_DPI_THRESHOLD = 128;
     AudioProcessorEditor* editor = new DexedAudioProcessorEditor (this);
     float scaleFactor = getDpiScaleFactor();
+    float maxFactor = DexedAudioProcessorEditor::getLargestScaleFactor();
 
-    if ( scaleFactor == -1 ) {
-        if ( Desktop::getInstance().getDisplays().getPrimaryDisplay()->dpi > HIGH_DPI_THRESHOLD ) {
-            scaleFactor = 1.5;
-        } else {
-            scaleFactor = 1.0;
-        }
+    if ( scaleFactor == -1 || scaleFactor > maxFactor ) {
+        scaleFactor = maxFactor;
     }
-    
-    const juce::Rectangle<int> rect(DexedAudioProcessorEditor::WINDOW_SIZE_X * scaleFactor,DexedAudioProcessorEditor::WINDOW_SIZE_Y * scaleFactor);
-    bool displayFound = false;
-    
-    // validate if there is really a display that can show the complete plugin size
-    for (auto& display : Desktop::getInstance().getDisplays().displays) {
-        if ( display.userArea.getHeight() > rect.getHeight() && display.userArea.getWidth() > rect.getWidth() )
-            displayFound = true;
-    }
-    
-    // no display found, scaling to default value	
-    if ( ! displayFound )
-        setDpiScaleFactor(1.0);
-    else 
-        setDpiScaleFactor(scaleFactor);
-        
+
+    setDpiScaleFactor(scaleFactor);
     return editor;
 }
 
